
class VehicleVisualizer:
    def __init__(self, vehicle: Vehicle):
        self.vehicle = vehicle

    def update_vehicle(self, vehicle: Vehicle):
        self.vehicle = vehicle


class InfrastructureVisualizer:
    def __init__(self):
        pass


class FrameVisualizer:
    def __init__(self):
        pass